package ru.dbotthepony.kommons.io

import ru.dbotthepony.kommons.math.RGBAColor
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.*
import kotlin.reflect.KClass

/**
 * Represents value which can be encoded onto or decoded from stream.
 */
interface StreamCodec<V> {
	fun read(stream: DataInputStream): V
	fun write(stream: DataOutputStream, value: V)

	/**
	 * Defensive copy
	 */
	fun copy(value: V): V

	/**
	 * Optional custom equality check
	 */
	fun compare(a: V, b: V): Boolean {
		return a == b
	}

	class Impl<V>(
		private val reader: (stream: DataInputStream) -> V,
		private val writer: (stream: DataOutputStream, value: V) -> Unit,
		private val copier: ((value: V) -> V) = { it },
		private val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
	) : StreamCodec<V> {
		constructor(
			reader: (stream: DataInputStream) -> V,
			payloadSize: Long,
			writer: (stream: DataOutputStream, value: V) -> Unit,
			copier: ((value: V) -> V) = { it },
			comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
		) : this({ stream -> reader.invoke(stream) }, writer, copier, comparator)

		val nullable = object : StreamCodec<V?> {
			override fun read(stream: DataInputStream): V? {
				return if (stream.read() == 0) null else reader.invoke(stream)
			}

			override fun write(stream: DataOutputStream, value: V?) {
				if (value === null) stream.write(0) else {
					stream.write(1)
					writer.invoke(stream, value)
				}
			}

			override fun copy(value: V?): V? {
				return if (value === null) null else copier.invoke(value)
			}

			override fun compare(a: V?, b: V?): Boolean {
				if (a === null && b === null) return true
				if (a === null || b === null) return false
				return comparator.invoke(a, b)
			}
		}

		override fun read(stream: DataInputStream): V {
			return reader.invoke(stream)
		}

		override fun write(stream: DataOutputStream, value: V) {
			writer.invoke(stream, value)
		}

		override fun copy(value: V): V {
			return copier.invoke(value)
		}

		override fun compare(a: V, b: V): Boolean {
			return comparator.invoke(a, b)
		}
	}

	class Collection<E, C : MutableCollection<E>>(val elementCodec: StreamCodec<E>, val collectionFactory: (Int) -> C) : StreamCodec<C> {
		override fun read(stream: DataInputStream): C {
			val size = stream.readVarInt()

			if (size <= 0) {
				return collectionFactory.invoke(0)
			}

			val collection = collectionFactory.invoke(size)

			for (i in 0 until size) {
				collection.add(elementCodec.read(stream))
			}

			return collection
		}

		override fun write(stream: DataOutputStream, value: C) {
			stream.writeVarInt(value.size)
			value.forEach { elementCodec.write(stream, it) }
		}

		override fun copy(value: C): C {
			val new = collectionFactory.invoke(value.size)
			value.forEach { new.add(elementCodec.copy(it)) }
			return new
		}
	}

	class Enum<V : kotlin.Enum<V>>(clazz: Class<out V>) : StreamCodec<V> {
		val clazz = searchClass(clazz)
		val values: List<V> = listOf(*this.clazz.enumConstants!!)
		val valuesMap = values.associateBy { it.name }

		override fun read(stream: DataInputStream): V {
			val id = stream.readVarInt()
			return values.getOrNull(id) ?: throw NoSuchElementException("No such enum with index $id")
		}

		override fun write(stream: DataOutputStream, value: V) {
			stream.writeVarInt(value.ordinal)
		}

		override fun copy(value: V): V {
			return value
		}

		override fun compare(a: V, b: V): Boolean {
			return a === b
		}

		companion object {
			/**
			 * FIXME: enums with abstract methods which get compiled to subclasses, whose DO NOT expose "parent's" enum constants array
			 *
			 * is there an already existing solution?
			 */
			fun <V : kotlin.Enum<V>> searchClass(clazz: Class<out V>): Class<out V> {
				var search: Class<*> = clazz

				while (search.enumConstants == null && search.superclass != null) {
					search = search.superclass
				}

				if (search.enumConstants == null) {
					throw ClassCastException("$clazz does not represent an enum or enum subclass")
				}

				return search as Class<out V>
			}
		}
	}

	class KOptional<V>(val codec: StreamCodec<V>) : StreamCodec<ru.dbotthepony.kommons.util.KOptional<V>> {
		override fun read(stream: DataInputStream): ru.dbotthepony.kommons.util.KOptional<V> {
			return stream.readKOptional(codec::read)
		}

		override fun write(stream: DataOutputStream, value: ru.dbotthepony.kommons.util.KOptional<V>) {
			stream.writeKOptional(value, codec::write)
		}

		override fun copy(value: ru.dbotthepony.kommons.util.KOptional<V>): ru.dbotthepony.kommons.util.KOptional<V> {
			return value.map(codec::copy)
		}
	}

	class Optional<V : Any>(val codec: StreamCodec<V>) : StreamCodec<java.util.Optional<V>> {
		override fun read(stream: DataInputStream): java.util.Optional<V> {
			return stream.readOptional(codec::read)
		}

		override fun write(stream: DataOutputStream, value: java.util.Optional<V>) {
			stream.writeOptional(value, codec::write)
		}

		override fun copy(value: java.util.Optional<V>): java.util.Optional<V> {
			return value.map(codec::copy)
		}
	}
}

val NullValueCodec = StreamCodec.Impl({ _ -> null }, { _, _ -> })
val BooleanValueCodec = StreamCodec.Impl(DataInputStream::readBoolean, 1L, DataOutputStream::writeBoolean)
val ByteValueCodec = StreamCodec.Impl(DataInputStream::readByte, 1L, { s, v -> s.writeByte(v.toInt()) })
val ShortValueCodec = StreamCodec.Impl(DataInputStream::readShort, 2L, { s, v -> s.writeShort(v.toInt()) })
val CharValueCodec = StreamCodec.Impl(DataInputStream::readChar, 2L, { s, v -> s.writeShort(v.code) })
val IntValueCodec = StreamCodec.Impl(DataInputStream::readInt, 4L, DataOutputStream::writeInt)
val LongValueCodec = StreamCodec.Impl(DataInputStream::readLong, 8L, DataOutputStream::writeLong)
val FloatValueCodec = StreamCodec.Impl(DataInputStream::readFloat, 4L, DataOutputStream::writeFloat)
val DoubleValueCodec = StreamCodec.Impl(DataInputStream::readDouble, 8L, DataOutputStream::writeDouble)
val BigDecimalValueCodec = StreamCodec.Impl(DataInputStream::readBigDecimal, DataOutputStream::writeBigDecimal)
val UUIDValueCodec = StreamCodec.Impl({ s -> UUID(s.readLong(), s.readLong()) }, { s, v -> s.writeLong(v.mostSignificantBits); s.writeLong(v.leastSignificantBits) })
val VarIntValueCodec = StreamCodec.Impl(DataInputStream::readSignedVarInt, DataOutputStream::writeSignedVarInt)
val VarLongValueCodec = StreamCodec.Impl(DataInputStream::readSignedVarLong, DataOutputStream::writeSignedVarLong)
val BinaryStringCodec = StreamCodec.Impl(DataInputStream::readBinaryString, DataOutputStream::writeBinaryString)

val RGBCodec = StreamCodec.Impl(
	{ s -> RGBAColor(s.readFloat(), s.readFloat(), s.readFloat()) },
	{ s, v -> s.writeFloat(v.red); s.writeFloat(v.green); s.writeFloat(v.blue) })

val RGBACodec = StreamCodec.Impl(
	{ s -> RGBAColor(s.readFloat(), s.readFloat(), s.readFloat(), s.readFloat()) },
	{ s, v -> s.writeFloat(v.red); s.writeFloat(v.green); s.writeFloat(v.blue); s.writeFloat(v.alpha) })

val OptionalBooleanValueCodec = StreamCodec.Optional(BooleanValueCodec)
val OptionalByteValueCodec = StreamCodec.Optional(ByteValueCodec)
val OptionalShortValueCodec = StreamCodec.Optional(ShortValueCodec)
val OptionalCharValueCodec = StreamCodec.Optional(CharValueCodec)
val OptionalIntValueCodec = StreamCodec.Optional(IntValueCodec)
val OptionalLongValueCodec = StreamCodec.Optional(LongValueCodec)
val OptionalFloatValueCodec = StreamCodec.Optional(FloatValueCodec)
val OptionalDoubleValueCodec = StreamCodec.Optional(DoubleValueCodec)
val OptionalBigDecimalValueCodec = StreamCodec.Optional(BigDecimalValueCodec)
val OptionalUUIDValueCodec = StreamCodec.Optional(UUIDValueCodec)
val OptionalVarIntValueCodec = StreamCodec.Optional(VarIntValueCodec)
val OptionalVarLongValueCodec = StreamCodec.Optional(VarLongValueCodec)
val OptionalBinaryStringCodec = StreamCodec.Optional(BinaryStringCodec)
val OptionalRGBCodec = StreamCodec.Optional(RGBCodec)
val OptionalRGBACodec = StreamCodec.Optional(RGBACodec)

val KOptionalBooleanValueCodec = StreamCodec.KOptional(BooleanValueCodec)
val KOptionalByteValueCodec = StreamCodec.KOptional(ByteValueCodec)
val KOptionalShortValueCodec = StreamCodec.KOptional(ShortValueCodec)
val KOptionalCharValueCodec = StreamCodec.KOptional(CharValueCodec)
val KOptionalIntValueCodec = StreamCodec.KOptional(IntValueCodec)
val KOptionalLongValueCodec = StreamCodec.KOptional(LongValueCodec)
val KOptionalFloatValueCodec = StreamCodec.KOptional(FloatValueCodec)
val KOptionalDoubleValueCodec = StreamCodec.KOptional(DoubleValueCodec)
val KOptionalBigDecimalValueCodec = StreamCodec.KOptional(BigDecimalValueCodec)
val KOptionalUUIDValueCodec = StreamCodec.KOptional(UUIDValueCodec)
val KOptionalVarIntValueCodec = StreamCodec.KOptional(VarIntValueCodec)
val KOptionalVarLongValueCodec = StreamCodec.KOptional(VarLongValueCodec)
val KOptionalBinaryStringCodec = StreamCodec.KOptional(BinaryStringCodec)
val KOptionalRGBCodec = StreamCodec.KOptional(RGBCodec)
val KOptionalRGBACodec = StreamCodec.KOptional(RGBACodec)

fun <E : Enum<E>> Class<E>.codec() = StreamCodec.Enum(this)
fun <E : Enum<E>> KClass<E>.codec() = StreamCodec.Enum(this.java)
