package ru.dbotthepony.kommons.util

import java.util.function.Supplier

fun <T> KOptional(value: T) = KOptional.of(value)
fun KOptional(value: Boolean) = KOptional.of(value)
fun KOptional(value: Boolean?) = KOptional.of(value)
fun <T> KOptional() = KOptional.empty<T>()

/**
 * [java.util.Optional] supporting nulls
 *
 * This is done for structures, where value can be absent,
 * or can be present, including literal "null" as possible present value,
 * in more elegant solution than handling nullable Optionals
 */
class KOptional<T> private constructor(private val _value: T, val isPresent: Boolean) : Supplier<T> {
	/**
	 * @throws NoSuchElementException
	 */
	val value: T get() {
		if (isPresent) {
			return _value
		}

		throw NoSuchElementException("No value is present")
	}

	inline val isEmpty: Boolean
		get() = !isPresent

	inline fun ifPresent(block: (T) -> Unit): KOptional<T> {
		if (isPresent) {
			block.invoke(value)
		}

		return this
	}

	inline fun ifNotPresent(block: () -> Unit): KOptional<T> {
		if (!isPresent) {
			block.invoke()
		}

		return this
	}

	inline fun <R> map(block: (T) -> R): KOptional<R> {
		if (isPresent) {
			return of(block.invoke(value))
		} else {
			return empty()
		}
	}

	inline fun <R> flatMap(block: (T) -> KOptional<R>): KOptional<R> {
		if (isPresent) {
			return block.invoke(value)
		} else {
			return empty()
		}
	}

	@Suppress("NOTHING_TO_INLINE")
	inline fun orElse(value: T): T {
		if (isPresent) {
			return this.value
		} else {
			return value
		}
	}

	inline fun orElse(value: () -> T): T {
		if (isPresent) {
			return this.value
		} else {
			return value.invoke()
		}
	}

	inline fun orThrow(supplier: () -> Throwable): T {
		if (isPresent) {
			return this.value
		} else {
			throw supplier.invoke()
		}
	}

	infix fun or(value: KOptional<T>): KOptional<T> {
		if (isPresent) {
			return this
		} else {
			return value
		}
	}

	fun valueEquals(other: Any?): Boolean {
		if (isPresent) {
			return _value == other
		}

		return false
	}

	override fun equals(other: Any?): Boolean {
		return this === other || other is KOptional<*> && isPresent == other.isPresent && _value == other._value
	}

	override fun hashCode(): Int {
		return _value.hashCode() + 43839429
	}

	override fun toString(): String {
		if (isPresent) {
			return "KOptional[value = $value]"
		} else {
			return "KOptional[empty]"
		}
	}

	override fun get(): T {
		return value
	}

	// gson hack since type token can't diff between nullable and non-null types
	@Deprecated("reflection class, don't use directly")
	class Nullable<T : Any?> private constructor()
	@Deprecated("reflection class, don't use directly")
	class NotNull<T : Any> private constructor()

	companion object {
		private val EMPTY = KOptional(null, false)
		private val NULL = KOptional(null, true)
		private val TRUE = KOptional(true, true)
		private val FALSE = KOptional(false, true)

		@JvmStatic
		fun <T> empty(): KOptional<T> {
			return EMPTY as KOptional<T>
		}

		@JvmStatic
		fun <T> of(value: T): KOptional<T> {
			if (value == null) {
				return NULL as KOptional<T>
			} else {
				return KOptional(value, true)
			}
		}

		@JvmStatic
		fun <T : Any> ofNullable(value: T?): KOptional<T> {
			if (value == null) {
				return EMPTY as KOptional<T>
			} else {
				return KOptional(value, true)
			}
		}

		@JvmStatic
		fun of(value: Boolean): KOptional<Boolean> {
			if (value) {
				return TRUE
			} else {
				return FALSE
			}
		}

		@JvmStatic
		fun of(value: Boolean?): KOptional<Boolean> {
			if (value == null) {
				return NULL as KOptional<Boolean>
			} else if (value) {
				return TRUE
			} else {
				return FALSE
			}
		}
	}
}
