package ru.dbotthepony.kommons.util

import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.concurrent.withLock

interface ValueObserver<V> : Listenable<V>, Supplier<V>, Observer {
	fun getAndObserve(): Pair<V, Boolean>

	/**
	 * Notifies listeners when value returned by [provider] changes.
	 *
	 * Observation happen on [get] call, or on [observe]/[getAndObserve].
	 *
	 * This implementation is not synchronized, calls to [get]/[observe]/[getAndObserve] from multiple threads
	 * will result in undefined behavior;
	 * For thread-safe implementation, use [SynchronizedImpl].
	 */
	class Impl<V>(private val provider: Supplier<V>) : ValueObserver<V> {
		private val subs = Listenable.Impl<V>()
		private var observed: Any? = Companion

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return subs.addListener(listener)
		}

		/**
		 * Returns last observed value or observes it if it were never observed before
		 */
		fun getObserved(): V {
			if (observed === Companion) {
				return get()
			}

			return observed as V
		}

		override fun get(): V {
			val get = provider.get()

			if (get != observed) {
				observed = get
				subs.accept(get)
			}

			return get
		}

		override fun observe(): Boolean {
			val get = provider.get()

			if (get != observed) {
				observed = get
				subs.accept(get)
				return true
			}

			return false
		}

		/**
		 * Returns pair representing value and if value changed since last observation.
		 */
		override fun getAndObserve(): Pair<V, Boolean> {
			val get = provider.get()

			if (get != observed) {
				observed = get
				subs.accept(get)
				return get to true
			}

			return get to false
		}

		private companion object
	}

	/**
	 * Notifies listeners when value returned by [provider] changes.
	 *
	 * Observation happen on [get] call, or on [observe]/[getAndObserve].
	 */
	class SynchronizedImpl<V>(private val provider: Supplier<V>) : ValueObserver<V> {
		private val subs = Listenable.Impl<V>()
		private val lock = ReentrantLock() // reentrant lock to avoid deadlock when doing observe -> listeners triggered -> listeners call observe
		private var observed: Any? = Companion

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return subs.addListener(listener)
		}

		/**
		 * Returns last observed value or observes it if it were never observed before
		 */
		fun getObserved(): V {
			if (observed === Companion) {
				return get()
			}

			return observed as V
		}

		override fun get(): V {
			val get = provider.get()

			if (get != observed) {
				lock.withLock {
					if (get != observed) {
						observed = get
						subs.accept(get)
					}
				}
			}

			return get
		}

		override fun observe(): Boolean {
			val get = provider.get()

			if (get != observed) {
				lock.withLock {
					if (get != observed) {
						observed = get
						subs.accept(get)
						return true
					}
				}
			}

			return false
		}

		/**
		 * Returns pair representing value and if value changed since last observation.
		 */
		override fun getAndObserve(): Pair<V, Boolean> {
			val get = provider.get()

			if (get != observed) {
				lock.withLock {
					if (get != observed) {
						observed = get
						subs.accept(get)
						return get to true
					}
				}
			}

			return get to false
		}

		private companion object
	}
}
